<?php
/*

MSN class ver 1.2 by Tommy Wu
License: GPL

You can find MSN protocol from this site: http://msnpiki.msnfanatic.com/index.php/Main_Page

This class support both MSNP15 and MSNP9 for send message. The PHP module needed:

  MSNP9: curl pcre
  MSNP15: curl pcre mhash mcrypt bcmath

Usually, this class will try to use MSNP15 if your system can support it, if your system can't support it,
it will switch to use MSNP9. But if you use MSNP9, it won't support OIM (Offline Messages).

Sameple Code:

$msn = new MSN;

if (!$msn->connect('YOUR_MSN_ID', 'YOUR_MSN_PASSWORD')) {
    echo "Error for connect to MSN network\n";
    echo "$msn->error\n";
    exit;
}

$msn->sendMessage('Now: '.strftime('%D %T')."\nTesting\nSecond Line\n\n\n\nand Empty Line",
                  array(
                    'somebody1@hotmail.com',
                    'somebody2@hotmail.com'
                       )
                 );
echo "Done!\n";
exit;

changelog:

1.2 2007/03/13
 ! fix the problem for 32bits platform can't get correct challenge code

1.1 2007/03/13
 + add timeout detect, make sure we can quit from loop when protocol changed or sometime wrong
 + allow save the debug information to file

1.0 2007/03/11
 = public release

*/

class MSN
{
	var $server	= 'messenger.hotmail.com';
	var $port = 1863;

    var $passport_url = 'https://login.live.com/RST.srf';
    var $protocol = 'MSNP15';
    var $buildver = '8.1.0178';
    var $prod_key = 'PK}_A_0N_K%O?A9S';
    var $prod_id = 'PROD0114ES4Z%Q5W';
    var $login_method = 'SSO';

    var $oim_url = 'https://ows.messenger.msn.com/OimWS/oim.asmx';
    var $oim_soap = 'http://messenger.live.com/ws/2006/09/oim/Store2';

    var $id;
    var $fp;
    var $error = '';

    var $authed = false;
    var $user = '';
    var $password = '';

    var $oim_try = 3;
    var $oim_ticket = '';

    var $debug = false;
    var $log_file = '';
    var $timeout = 15;

    var $sb;
	var $font_fn = 'Arial';
	var $font_co = '333333';
	var $font_ef = '';

    function MSN($protocol = '', $debug = false, $timeout = 15)
    {
        if (is_string($debug) && $debug !== '') {
            $this->debug = true;
            $this->log_file = $debug;
        }
        else
            $this->debug = $debug;
        $this->timeout = $timeout;
        // check support
        if (!function_exists('curl_init')) die("We need curl module!\n");
        if (!function_exists('preg_match')) die("We need pcre module!\n");

        if ($protocol != 'MSNP9' && $protocol != 'MSNP15')
            $protocol = '';

        if ($protocol != 'MSNP9' && !function_exists('mhash')) {
            if ($protocol == 'MSNP15') die("We need mhash module for MSNP15!\n");
            $protocol = 'MSNP9';
        }
        if ($protocol != 'MSNP9' && !function_exists('mcrypt_cbc')) {
            if ($protocol == 'MSNP15') die("We need mcrypt module for MSNP15!\n");
            $protocol = 'MSNP9';
        }
        if ($protocol != 'MSNP9' && !function_exists('bcmod')) {
            if ($protocol == 'MSNP15') die("We need bcmath module for MSNP15!\n");
            $protocol = 'MSNP9';
        }
        if ($protocol == 'MSNP9') {
            $this->protocol = 'MSNP9';
            $this->passport_url = 'https://nexus.passport.com/rdr/pprdr.asp';
            $this->buildver = '6.0.0602';
            $this->prod_key = 'Q1P7W2E4J9R8U3S5';
            $this->prod_id = 'msmsgs@msnmsgr.com';
            $this->login_method = 'TWN';
        }
        else {
            $this->protocol = 'MSNP15';
            $this->passport_url = 'https://login.live.com/RST.srf';
            $this->buildver = '8.1.0178';
            $this->prod_key = 'PK}_A_0N_K%O?A9S';
            $this->prod_id = 'PROD0114ES4Z%Q5W';
            $this->login_method = 'SSO';

            $this->oim_url = 'https://ows.messenger.msn.com/OimWS/oim.asmx';
            $this->oim_soap = 'http://messenger.live.com/ws/2006/09/oim/Store2';
        }
        return;
    }

    function get_passport_ticket($policy)
    {
        $user = htmlspecialchars($this->user);
        $password = htmlspecialchars($this->password);

        $XML = '<?xml version="1.0" encoding="UTF-8"?>
<Envelope xmlns="http://schemas.xmlsoap.org/soap/envelope/"
          xmlns:wsse="http://schemas.xmlsoap.org/ws/2003/06/secext"
          xmlns:saml="urn:oasis:names:tc:SAML:1.0:assertion"
          xmlns:wsp="http://schemas.xmlsoap.org/ws/2002/12/policy"
          xmlns:wsu="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd"
          xmlns:wsa="http://schemas.xmlsoap.org/ws/2004/03/addressing"
          xmlns:wssc="http://schemas.xmlsoap.org/ws/2004/04/sc"
          xmlns:wst="http://schemas.xmlsoap.org/ws/2004/04/trust">
<Header>
  <ps:AuthInfo xmlns:ps="http://schemas.microsoft.com/Passport/SoapServices/PPCRL" Id="PPAuthInfo">
    <ps:HostingApp>{7108E71A-9926-4FCB-BCC9-9A9D3F32E423}</ps:HostingApp>
    <ps:BinaryVersion>4</ps:BinaryVersion>
    <ps:UIVersion>1</ps:UIVersion>
    <ps:Cookies></ps:Cookies>
    <ps:RequestParams>AQAAAAIAAABsYwQAAAAxMDMz</ps:RequestParams>
  </ps:AuthInfo>
  <wsse:Security>
    <wsse:UsernameToken Id="user">
      <wsse:Username>'.$user.'</wsse:Username>
      <wsse:Password>'.$password.'</wsse:Password>
    </wsse:UsernameToken>
  </wsse:Security>
</Header>
<Body>
  <ps:RequestMultipleSecurityTokens xmlns:ps="http://schemas.microsoft.com/Passport/SoapServices/PPCRL" Id="RSTS">
    <wst:RequestSecurityToken Id="RST0">
      <wst:RequestType>http://schemas.xmlsoap.org/ws/2004/04/security/trust/Issue</wst:RequestType>
      <wsp:AppliesTo>
        <wsa:EndpointReference>
          <wsa:Address>http://Passport.NET/tb</wsa:Address>
        </wsa:EndpointReference>
      </wsp:AppliesTo>
    </wst:RequestSecurityToken>
    <wst:RequestSecurityToken Id="RST1">
      <wst:RequestType>http://schemas.xmlsoap.org/ws/2004/04/security/trust/Issue</wst:RequestType>
      <wsp:AppliesTo>
        <wsa:EndpointReference>
          <wsa:Address>messengerclear.live.com</wsa:Address>
        </wsa:EndpointReference>
      </wsp:AppliesTo>
      <wsse:PolicyReference URI="'.$policy.'"></wsse:PolicyReference>
    </wst:RequestSecurityToken>
    <wst:RequestSecurityToken Id="RST2">
      <wst:RequestType>http://schemas.xmlsoap.org/ws/2004/04/security/trust/Issue</wst:RequestType>
      <wsp:AppliesTo>
        <wsa:EndpointReference>
          <wsa:Address>messengersecure.live.com</wsa:Address>
        </wsa:EndpointReference>
      </wsp:AppliesTo>
      <wsse:PolicyReference URI="MBI_SSL"></wsse:PolicyReference>
    </wst:RequestSecurityToken>
  </ps:RequestMultipleSecurityTokens>
</Body>
</Envelope>';

        $this->debug_message("*** Sending SOAP:\n$XML");
        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $this->passport_url);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($curl, CURLOPT_FOLLOWLOCATION, 1);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, 0);
        curl_setopt($curl, CURLOPT_POST, 1);
        curl_setopt($curl, CURLOPT_POSTFIELDS, $XML);
        $data = curl_exec($curl);
        curl_close($curl);
        $this->debug_message("*** Get Result:\n$data");

        // we need ticket and secret code
        // <wsse:BinarySecurityToken Id="Compact1">t=tick&p=</wsse:BinarySecurityToken>
        // ...
        // <wst:BinarySecret>binary secret</wst:BinarySecret>
        preg_match("#<wsse\:BinarySecurityToken Id=\"Compact1\">(.*)</wsse:BinarySecurityToken>(.*)<wst\:BinarySecret>(.*)</wst\:BinarySecret>(.*)<wsse\:BinarySecurityToken Id=\"Compact2\">(.*)</wsse:BinarySecurityToken>#",
                   $data, $matches);

        // no ticket found!
        if (count($matches) <= 0) {
            $this->debug_message("*** Can't get passport ticket!");
            return false;
        }

        // yes, we get ticket
        return array(
                    'ticket' => html_entity_decode($matches[1]),
                    'secret' => html_entity_decode($matches[3]),
                    'oim_ticket' => html_entity_decode($matches[5])
                    );
    }

	function get_tweener_passport_ticket($nonce)
	{
        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $this->passport_url);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($curl, CURLOPT_FOLLOWLOCATION, 1);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, 0);
        curl_setopt($curl, CURLOPT_HEADER, 1);
        curl_setopt($curl, CURLOPT_NOBODY, 1);
        $data = curl_exec($curl);
        curl_close($curl);
        $this->debug_message("*** Get Result:\n$data");

        // we need login URL
        // DALogin=xxx
        preg_match('/DALogin=(.*?),/', $data, $matches);

        // no URL found!
        if (count($matches) <= 0) {
            $this->debug_message("*** Can't get passport's URL!");
            return false;
        }

        $url = 'https://'.$matches[1];

        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_HTTPHEADER, array(
                        'Authorization: Passport1.4 OrgVerb=GET,OrgURL=http%3A%2F%2Fmessenger%2Emsn%2Ecom,sign-in='.$this->user.',pwd='.$this->password.','.$nonce,
                        'Host: login.passport.com'
                        ));
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($curl, CURLOPT_FOLLOWLOCATION, 1);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, 0);
        curl_setopt($curl, CURLOPT_HEADER, 1);
        curl_setopt($curl, CURLOPT_NOBODY, 1);
        $data = curl_exec($curl);
        curl_close($curl);
        $this->debug_message("*** Get Result:\n$data");

        // we need ticket
        // from-PP=xxx
        preg_match("/from-PP='(.*?)'/", $data, $matches);

        // no URL found!
        if (count($matches) <= 0) {
            $this->debug_message("*** Can't get passport's ticket!");
            return false;
        }
        return $matches[1];
	}

	function connect($user, $password)
	{
		$this->id = 1;
        $this->fp = @fsockopen($this->server, $this->port, $errno, $errstr, 5);
        if (!$this->fp) {
            $this->error = "Can't connect to $this->server:$this->port, error => $errno, $errstr";
            return false;
        }

        // MSNP9
        // NS: >> VER {id} MSNP9 CVR0
        // MSNP15
        // NS: >>> VER {id} MSNP15 CVR0
        $this->writeln("VER $this->id $this->protocol CVR0");

        $start_tm = time();
        while (!feof($this->fp)) {
            $data = $this->readln();
            if ($this->timeout > 0) {
                $now_tm = time();
                $used_time = ($now_tm >= $start_tm) ? $now_tm - $start_tm : $now_tm;
                if ($used_time > $this->timeout) {
                    // logout now
                    // NS: >>> OUT
                    $this->writeln("OUT");
                    fclose($this->fp);
                    $this->error = 'Timeout, maybe protocol changed!';
                    $this->debug_message("*** $this->error");
                    return false;
                }
            }
            // no data?
            if ($data === false) continue;
            $code = substr($data, 0, 3);
            switch ($code) {
                case 'VER':
                    // MSNP9
                    // NS: <<< VER {id} MSNP9 CVR0
                    // NS: >>> CVR {id} 0x0409 winnt 5.1 i386 MSMSGS 6.0.0602 msmsgs {user}
                    // MSNP15
                    // NS: <<< VER {id} MSNP15 CVR0
                    // NS: >>> CVR {id} 0x0409 winnt 5.1 i386 MSMSGS 8.1.0178 msmsgs {user}
                    $this->writeln("CVR $this->id 0x0409 winnt 5.1 i386 MSMSGS $this->buildver msmsgs $user");
                    break;

                case 'CVR':
                    // MSNP9
                    // NS: <<< CVR {id} {ver_list} {download_serve} ....
                    // NS: >>> USR {id} TWN I {user}
                    // MSNP15
                    // NS: <<< CVR {id} {ver_list} {download_serve} ....
                    // NS: >>> USR {id} SSO I {user}
                    $this->writeln("USR $this->id $this->login_method I $user");
                    break;

                case 'USR':
                    // already login for passport site, finish the login process now.
                    // NS: <<< USR {id} OK {user} {verify} 0
                    if ($this->authed) return true;

                    $this->user = $user;
                    $this->password = urlencode($password);

                    if ($this->protocol == 'MSNP15') {
                        // NS: <<< USR {id} SSO S {policy} {nonce}
                        list(/* USR */, /* id */, /* SSO */, /* S */, $policy, $nonce,) = explode(' ', $data);

                        $result = $this->get_passport_ticket($policy);
                        if (!$result || !is_array($result)) {
                            // logout now
                            // NS: >>> OUT
                            $this->writeln("OUT");
                            fclose($this->fp);
                            $this->error = 'Passport authenticated fail!';
                            $this->debug_message("*** $this->error");
                            return false;
                        }

                        $ticket = $result['ticket'];
                        $secret = $result['secret'];
                        $this->oim_ticket = $result['oim_ticket'];

                        $login_code = $this->generateLoginBLOB($secret, $nonce);

                        // NS: >>> USR {id} SSO S {ticket} {login_code}
                        $this->writeln("USR $this->id $this->login_method S $ticket $login_code");
                    }
                    else {
                        // NS: <<< USR {id} TWN S {nonce}
                        list(/* USR */, /* id */, /* TWN */, /* S */, $nonce,) = explode(' ', $data);

                        $ticket = $this->get_tweener_passport_ticket($nonce);
                        if (!$ticket) {
                            // logout now
                            // NS: >>> OUT
                            $this->writeln("OUT");
                            fclose($this->fp);
                            $this->error = 'Passport authenticated fail!';
                            $this->debug_message("*** $this->error");
                            return false;
                        }

                        // NS: >>> USR {id} TWN S {ticket}
                        $this->writeln("USR $this->id $this->login_method S $ticket");
                    }
                    $this->authed = true;
                    break;

                case 'XFR':
                    // main login server will redirect to anther NS after USR command
                    // MSNP9
                    // NS: <<< XFR {id} NS {server} 0 {server}
                    // MSNP15
                    // NS: <<< XFR {id} NS {server} U D
                    list(/* XFR */, /* id */, /* NS */, $server, /* ... */) = explode(' ', $data);
                    list($ip, $port) = explode(':', $server);
                    // this connection will close after XFR
                    fclose($this->fp);

                    $this->fp = @fsockopen($ip, $port, $errno, $errstr, 5);
                    if (!$this->fp) {
                        $this->error = "Can't connect to $ip:$port, error => $errno, $errstr";
                        $this->debug_message("*** $this->error");
                        return false;
                    }

                    // MSNP9
                    // NS: >> VER {id} MSNP9 CVR0
                    // MSNP15
                    // NS: >>> VER {id} MSNP15 CVR0
                    $this->writeln("VER $this->id $this->protocol CVR0");
                    break;

                case 'GCF':
                    // return some policy data after 'USR {id} SSO I {user}' command
                    // NS: <<< GCF 0 {size}
                    list(/* GCF */, /* 0 */, $size,) = explode(' ', $data);
                    // we don't need the data, just read it and drop
                    $this->readdata($size);
                    break;

                default:
                    // we'll quit if got any error
                    if (is_numeric($code)) {
                        // logout now
                        // NS: >>> OUT
                        $this->writeln("OUT");
                        fclose($this->fp);
                        $this->error = "Error code: $code, please check the detail information from: http://msnpiki.msnfanatic.com/index.php/Reference:Error_List";
                        $this->debug_message("*** $this->error");
                        return false;
                    }
                    // unknown response from server, just ignore it
					break;
            }
        }
        // never goto here
    }

    function derive_key($key, $magic)
    {
        $hash1 = mhash(MHASH_SHA1, $magic, $key);
        $hash2 = mhash(MHASH_SHA1, $hash1.$magic, $key);
        $hash3 = mhash(MHASH_SHA1, $hash1, $key);
        $hash4 = mhash(MHASH_SHA1, $hash3.$magic, $key);
        return $hash2.substr($hash4, 0, 4);
    }

    function generateLoginBLOB($key, $challenge)
    {
        $key1 = base64_decode($key);
        $key2 = $this->derive_key($key1, 'WS-SecureConversationSESSION KEY HASH');
        $key3 = $this->derive_key($key1, 'WS-SecureConversationSESSION KEY ENCRYPTION');

        // get hash of challenge using key2
        $hash = mhash(MHASH_SHA1, $challenge, $key2);

        // get 8 bytes random data
        $iv = substr(base64_encode(rand(1000,9999).rand(1000,9999)), 2, 8);

        $cipher = mcrypt_cbc(MCRYPT_3DES, $key3, $challenge."\x08\x08\x08\x08\x08\x08\x08\x08", MCRYPT_ENCRYPT, $iv);

        $blob = pack('LLLLLLL', 28, 1, 0x6603, 0x8004, 8, 20, 72);
        $blob .= $iv;
        $blob .= $hash;
        $blob .= $cipher;

        return base64_encode($blob);
    }

	function sendMessage($sMessage, $aTo)
	{
        if ($this->protocol == 'MSNP9') {
            // we need to send SYN command for MSNP9
            // NS: >>> SYN {id} 0
            $this->writeln("SYN $this->id 0");
        }
        $quit = false;
        $online_cnt = 0;
        $offline_cnt = 0;
        $start_tm = time();
		while (!feof($this->fp)) {
            if ($quit) break;
			$data = $this->readln();
            if ($this->timeout > 0) {
                $now_tm = time();
                $used_time = ($now_tm >= $start_tm) ? $now_tm - $start_tm : $now_tm;
                if ($used_time > $this->timeout) {
                    $this->error = 'Timeout, maybe protocol changed!';
                    $this->debug_message("*** $this->error");
                    break;
                }
            }
            // no data ?
            if ($data === false) continue;
            $code = substr($data, 0, 3);

            switch ($code) {
                case 'SBS':
                    // after 'USR {id} OK {user} {verify} 0' response, the server will send SBS and profile to us
                    // NS: <<< SBS 0 null
                    // we don't need profile data, so just ignore it
                    // change our status to online first
                    // NS: >>> CHG {id} {status}
                    $this->writeln("CHG $this->id NLN");
                    break;

                case 'MSG':
                    // randomly, we get MSG notification from server
                    // NS: <<< MSG Hotmail Hotmail {size}
                    list(/* MSG */, /* Hotmail */, /* Hotmail */, $size,) = explode(' ', $data);
                    // we don't need the notification data, so just ignore it
                    if (is_numeric($size))
                        $this->readdata($size);
                    break;

                case 'CHL':
                    // randomly, we'll get challenge from server
                    // NS: <<< CHL 0 {code}
                    list(/* CHL */, /* 0 */, $chl_code,) = explode(' ', $data);
                    $fingerprint = $this->getChallenge($chl_code);
                    // NS: >>> QRY {id} {product_id} 32
                    // NS: >>> fingerprint
                    $this->writeln("QRY $this->id $this->prod_id 32");
                    $this->writedata($fingerprint);
                    break;

                case 'SYN':
                    if ($this->protocol != 'MSNP9') break;
                    // NS: <<< SYN 8 1 2 1
                    // ignore it
                    // change our status to online first
                    // NS: >>> CHG {id} {status}
                    $this->writeln("CHG $this->id NLN");
                    break;

                case 'CHG':
                    // NS: <<< CHG {id} {status} {code}
                    // ignore it

                    // if message is empty or To list is empty, just quit
                    if (count($aTo) == 0 || $sMessage === '') {
                        $quit = true;
                        break;
                    }
                    // okay, try to ask a switchboard (SB) for sending message
                    // NS: >>> XFR {id} SB
                    $this->writeln("XFR $this->id SB");
                    break;

                case 'XFR':
                    // NS: <<< XFR {id} SB {server} CKI {cki} U messenger.msn.com 0
                    list(/* XFR */, /* {id} */, /* SB */, $server, /* CKI */, $cki_code, /* ... */) = explode(' ', $data);
                    list($ip, $port) = explode(':', $server);
                    $result = $this->switchboard_control($ip, $port, $cki_code, $aTo, $sMessage);
                    if (!$result || !is_array($result)) {
                        // error for switchboard, quit
                        $quit = true;
                        break;
                    }
                    // okay, process offline user
                    if ($this->protocol == 'MSNP9') {
                        // MSNP9 don't support OIM
                        $this->debug_message("*** MSNP9 don't support OIM, so we ignore offline user!");
                        $quit = true;
                        break;
                    }
                    $lockkey = '';
                    foreach ($result['offline'] as $to) {
                        for ($i = 0; $i < $this->oim_try; $i++) {
                            $chl_code = $this->sendOIM($to, $sMessage, $lockkey);
                            if ($chl_code === false) {
                                // finished
                                break;
                            }
                            // need challenge lockkey
                            $lockkey = $this->getChallenge($chl_code);
                        }
                    }
                    $quit = true;
                    break;

                default:
                    if (is_numeric($code)) {
                        $this->error = "Error code: $code, please check the detail information from: http://msnpiki.msnfanatic.com/index.php/Reference:Error_List";
                        $this->debug_message("*** NS: $this->error");
                    }
                    break;
            }
		}
        // logout now
        // NS: >>> OUT
        $this->writeln("OUT");
        fclose($this->fp);
        return array(
                        'online' => $online_cnt,
                        'offline' => $offline_cnt
                    );
	}

    function getChallenge($code)
    {
        if ($this->protocol == 'MSNP9') {
            // simple challenge for MSNP9
            return md5($code.$this->prod_key);
        }
        // MSNP15
        // http://msnpiki.msnfanatic.com/index.php/MSNP11:Challenges
        // Step 1: The MD5 Hash
        $md5Hash = md5($code.$this->prod_key);
        $aMD5 = explode("\0", chunk_split($md5Hash, 8, "\0"));
        for ($i = 0; $i < 4; $i++) {
            $aMD5[$i] = implode('', array_reverse(explode("\0", chunk_split($aMD5[$i], 2, "\0"))));
            $aMD5[$i] = (0 + base_convert($aMD5[$i], 16, 10)) & 0x7FFFFFFF;
        }

        // Step 2: A new string
        $chl_id = $code.$this->prod_id;
        $chl_id .= str_repeat('0', 8 - (strlen($chl_id) % 8));

        $aID = explode("\0", substr(chunk_split($chl_id, 4, "\0"), 0, -1));
        for ($i = 0; $i < count($aID); $i++) {
            $aID[$i] = implode('', array_reverse(explode("\0", chunk_split($aID[$i], 1, "\0"))));
            $aID[$i] = 0 + base_convert(bin2hex($aID[$i]), 16, 10);
        }

        // Step 3: The 64 bit key
        $magic_num = 0x0E79A9C1;
        $str7f = 0x7FFFFFFF;
        $high = 0;
        $low = 0;
        for ($i = 0; $i < count($aID); $i += 2) {
            $temp = $aID[$i];
            $temp = bcmod(bcmul($magic_num, $temp), $str7f);
            $temp = bcadd($temp, $high);
            $temp = bcadd(bcmul($aMD5[0], $temp), $aMD5[1]);
            $temp = bcmod($temp, $str7f);

            $high = $aID[$i+1];
            $high = bcmod(bcadd($high, $temp), $str7f);
            $high = bcadd(bcmul($aMD5[2], $high), $aMD5[3]);
            $high = bcmod($high, $str7f);

            $low = bcadd(bcadd($low, $high), $temp);
        }

        $high = bcmod(bcadd($high, $aMD5[1]), $str7f);
        $low = bcmod(bcadd($low, $aMD5[3]), $str7f);

        $new_high = bcmul($high & 0xFF, 0x1000000);
        $new_high = bcadd($new_high, bcmul($high & 0xFF00, 0x100));
        $new_high = bcadd($new_high, bcdiv($high & 0xFF0000, 0x100));
        $new_high = bcadd($new_high, bcdiv($high & 0xFF000000, 0x1000000));
        // we need integer here
        $high = 0+$new_high;

        $new_low = bcmul($low & 0xFF, 0x1000000);
        $new_low = bcadd($new_low, bcmul($low & 0xFF00, 0x100));
        $new_low = bcadd($new_low, bcdiv($low & 0xFF0000, 0x100));
        $new_low = bcadd($new_low, bcdiv($low & 0xFF000000, 0x1000000));
        // we need integer here
        $low = 0+$new_low;

        // we just use 32 bits integer, don't need the key, just high/low
        // $key = bcadd(bcmul($high, 0x100000000), $low);

        // Step 4: Using the key
        $md5Hash = md5($code.$this->prod_key);
        $aHash = explode("\0", chunk_split($md5Hash, 8, "\0"));

        $hash = '';
        $hash .= sprintf("%08x", (0 + base_convert($aHash[0], 16, 10)) ^ $high);
        $hash .= sprintf("%08x", (0 + base_convert($aHash[1], 16, 10)) ^ $low);
        $hash .= sprintf("%08x", (0 + base_convert($aHash[2], 16, 10)) ^ $high);
        $hash .= sprintf("%08x", (0 + base_convert($aHash[3], 16, 10)) ^ $low);

        return $hash;
    }

    function switchboard_control($ip, $port, $cki_code, $aTo, $sMessage)
    {
        $this->debug_message("*** SB: try to connect to switchboard server $ip:$port");
        $this->sb = @fsockopen($ip, $port, $errno, $errstr, 5);
        if (!$this->sb) {
            $this->error = "SB: Can't connect to $ip:$port, error => $errno, $errstr";
            $this->debug_message("*** $this->error");
            return false;
        }

        if (is_array($aTo))
            $aToUsers = $aTo;
        else
            $aToUsers[0] = $aTo;

        $aOfflineUsers = array();
        $total_cnt = count($aToUsers);

        // SB: >>> USR {id} {user} {cki}
        $this->sb_writeln("USR $this->id $this->user $cki_code");

        $sent = false;
        $cal_cnt = 0;
        $offline_cnt = 0;
        $start_tm = time();
		while (!feof($this->sb)) {
            if ($sent) break;
			$data = $this->sb_readln();
            if ($this->timeout > 0) {
                $now_tm = time();
                $used_time = ($now_tm >= $start_tm) ? $now_tm - $start_tm : $now_tm;
                if ($used_time > $this->timeout) {
                    $this->error = 'Timeout, maybe protocol changed!';
                    $this->debug_message("*** $this->error");
                    break;
                }
            }
            if ($data === false) continue;
            $code = substr($data, 0, 3);
            switch($code) {
                case 'USR':
                    // SB: <<< USR {id} OK {user} {alias}
                    // we don't need the data, just ignore it
                    $user = $aToUsers[$cal_cnt++];
                    // request user to join this switchboard
                    // SB: >>> CAL {id} {user}
                    $this->sb_writeln("CAL $this->id $user");
                    break;

                case 'CAL':
                    // SB: <<< CAL {id} RINGING {?}
                    // we don't need this, just ignore, and wait for other response
                    break;

                case '217':
                    // SB: <<< 217 {id}
                    // if user isn't online or no such user, we will get 217.
                    // switchboard can't send message if he isn't online
                    $aOfflineUsers[$offline_cnt++] = $user;
                    // continue process with JOI here
                case 'JOI':
                    // SB: <<< JOI {user} {alias}
                    // someone join us
                    // we don't need the data, just ignore it
                    // try to call others if we've
                    if ($cal_cnt < $total_cnt) {
                        $user = $aToUsers[$cal_cnt++];
                        // request user to join this switchboard
                        // SB: >>> CAL {id} {user}
                        $this->sb_writeln("CAL $this->id $user");
                        break;
                    }
                    // no more user here
                    // check, all users are offline
                    if ($cal_cnt == $offline_cnt) {
                        $this->debug_message("*** SB: no any use online! skip to send message!");
                        $sent = true;
                        break;
                    }
                    $message = "MIME-Version: 1.0\r\nContent-Type: text/plain; charset=UTF-8\r\nX-MMS-IM-Format: FN=$this->font_fn; EF=$this->font_ef; CO=$this->font_co; CS=0; PF=22\r\n\r\n$sMessage";
                    $len = strlen($message);
                    $this->sb_writeln("MSG 20 N $len");
                    $this->sb_writedata($message);
                    $sent = true;
                    break;

                default:
                    if (is_numeric($code)) {
                        $this->error = "Error code: $code, please check the detail information from: http://msnpiki.msnfanatic.com/index.php/Reference:Error_List";
                        $this->debug_message("*** SB: $this->error");
                    }
                    break;
            }
        }
        $this->sb_writeln("OUT");
        @fclose($this->sb);
        if (count($aOfflineUsers) > 0) {
            return array('offline' => $aOfflineUsers);
        }
        return false;
    }


    function sendOIM($to, $sMessage, $lockkey)
    {
        // support OIM for MSNP13 or later (we using MSNP15)
        if ($this->oim_url == '') return false;
        $XML = '<?xml version="1.0" encoding="utf-8"?>
<soap:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
               xmlns:xsd="http://www.w3.org/2001/XMLSchema"
               xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/">
<soap:Header>
  <From memberName="'.$this->user.'"
        friendlyName="=?utf-8?B?'.base64_encode($this->user).'?="
        xml:lang="zh-TW"
        proxy="MSNMSGR"
        xmlns="http://messenger.msn.com/ws/2004/09/oim/"
        msnpVer="'.$this->protocol.'"
        buildVer="'.$this->buildver.'"/>
  <To memberName="'.$to.'" xmlns="http://messenger.msn.com/ws/2004/09/oim/"/>
  <Ticket passport="'.htmlspecialchars($this->oim_ticket).'"
          appid="'.$this->prod_id.'"
          lockkey="'.$lockkey.'"
          xmlns="http://messenger.msn.com/ws/2004/09/oim/"/>
  <Sequence xmlns="http://schemas.xmlsoap.org/ws/2003/03/rm">
    <Identifier xmlns="http://schemas.xmlsoap.org/ws/2002/07/utility">http://messenger.msn.com</Identifier>
    <MessageNumber>1</MessageNumber>
  </Sequence>
</soap:Header>
<soap:Body>
  <MessageType xmlns="http://messenger.msn.com/ws/2004/09/oim/">text</MessageType>
  <Content xmlns="http://messenger.msn.com/ws/2004/09/oim/">MIME-Version: 1.0
Content-Type: text/plain; charset=UTF-8
Content-Transfer-Encoding: base64
X-OIM-Message-Type: OfflineMessage
X-OIM-Run-Id: {DAB68CFA-38C9-449B-945E-38AFA51E50A7}
X-OIM-Sequence-Num: 1

'.chunk_split(base64_encode($sMessage)).'
  </Content>
</soap:Body>
</soap:Envelope>';

        $header_array = array(
                                'SOAPAction: '.$this->oim_soap,
                                'Content-Type: text/xml',
                                'User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; Messenger '.$this->buildver.')'
                            );

        $this->debug_message("*** Sending SOAP:\n$XML");
        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $this->oim_url);
        curl_setopt($curl, CURLOPT_HTTPHEADER, $header_array);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($curl, CURLOPT_FOLLOWLOCATION, 1);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, 0);
        curl_setopt($curl, CURLOPT_HEADER, 1);
        curl_setopt($curl, CURLOPT_POST, 1);
        curl_setopt($curl, CURLOPT_POSTFIELDS, $XML);
        $data = curl_exec($curl);
        curl_close($curl);
        $this->debug_message("*** Get Result:\n$data");

        // the lockkey is invalid, authenticated fail, we need challenge it again
        // <LockKeyChallenge xmlns="http://messenger.msn.com/ws/2004/09/oim/">364763969</LockKeyChallenge>
		preg_match("#<LockKeyChallenge (.*)>(.*)</LockKeyChallenge>#", $data, $matches);
        // no challenge, assume it work
        if (count($matches) == 0) {
            //<faultcode xmlns:q0="http://messenger.msn.com/ws/2004/09/oim/">q0:AuthenticationFailed</faultcode>
            preg_match("#<faultcode (.*)>(.*)</faultcode>#", $data, $matches);
            if (count($matches) == 0) {
                // no error, we assume the OIM is sent
                $this->debug_message("*** OIM sent for $to");
                return false;
            }
            $err_code = $matches[2];
            //<faultstring>Exception of type 'System.Web.Services.Protocols.SoapException' was thrown.</faultstring>
            preg_match("#<faultstring>(.*)</faultstring>#", $data, $matches);
            if (count($matches) > 0)
                $err_msg = $matches[1];
            else
                $err_msg = '';
            $this->debug_message("*** OIM failed for $to");
            $this->debug_message("*** OIM Error code: $err_code");
            $this->debug_message("*** OIM Error Message: $err_msg");
            return false;
        }
        // yes, we get new LockKeyChallenge
        $challenge = $matches[2];
        $this->debug_message("*** OIM need new challenge ($challenge) for $to");
        return $challenge;
    }

    // read data for specified size
    function readdata($size)
    {
        $data = '';
        $count = 0;
        while (!feof($this->fp)) {
            $buf = @fread($this->fp, $size - $count);
            $data .= $buf;
            $count += strlen($buf);
            if ($count >= $size) break;
        }
        $this->debug_message("NS: data ($size/$count) <<<\n$data");
        return $data;
    }

    // read one line
	function readln()
	{
        $data = @fgets($this->fp, 4096);
        if ($data !== false) {
            $data = trim($data);
            $this->debug_message("NS: <<< $data");
        }
        return $data;
	}

    // write to server, append \r\n, also increase id
	function writeln($data)
	{
		fwrite($this->fp, $data."\r\n");
        $this->debug_message("NS: >>> $data");
        $this->id++;
        return;
	}

    // write data to server
	function writedata($data)
	{
		fwrite($this->fp, $data);
        $this->debug_message("NS: >>> $data");
        return;
	}

    // read one line for SB
	function sb_readln()
	{
        $data = @fgets($this->sb, 4096);
        if ($data !== false) {
            $data = trim($data);
            $this->debug_message("SB: <<< $data");
        }
        return $data;
	}

    // write to server for SB, append \r\n, also increase id
    // switchboard server only accept \r\n, it will lost connection if just \n only
	function sb_writeln($data)
	{
		fwrite($this->sb, $data."\r\n");
        $this->debug_message("SB: >>> $data");
        $this->id++;
        return;
	}

    // write data to server
	function sb_writedata($data)
	{
		fwrite($this->sb, $data);
        $this->debug_message("SB: >>> $data");
        return;
	}

    // show debug information
    function debug_message($str)
    {
        if (!$this->debug) return;
        if ($this->log_file !== '') {
            $fp = fopen($this->log_file, 'at');
            if ($fp) {
                fputs($fp, strftime('%D %T').' ['.posix_getpid().'] '.$str."\n");
                fclose($fp);
                return;
            }
            // still show debug information, if we can't open log_file
        }
        echo $str."\n";
        return;
    }
}


?>
